# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
from mo_streams._utils import Stream


class EmptyStream(Stream):
    def __getattr__(self, item):
        return self

    def __call__(self, *args, **kwargs):
        return self

    def to_dict(self):
        return {}

    def to_list(self):
        return []

    def to_bytes(self):
        return b""

    def to_str(self):
        return ""



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from mo_imports import export, expect

from mo_streams.type_utils import CallableTyper

Typer = expect("Typer")


def parse(type_desc):
    """
    THIS FUNCTION IS INCOMPLETE
    """
    if isinstance(type_desc, type):
        # SOME TYPE ANNOTATIONS ARE ACTUAL TYPES, NOT STRINGS
        return CallableTyper(type_=type_desc)

    types = [
        clean for t in type_desc.split("|") for clean in [t.strip()] if clean != "None"
    ]

    if len(types) == 1:
        if types[0] == "str":
            return CallableTyper(type_=str)

    raise NotImplementedError()


export("mo_streams.type_utils", parse)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#
import inspect

from mo_imports import expect, export
from mo_logs import logger

from mo_streams._utils import arg_spec

parse, ANNOTATIONS = expect("parse", "ANNOTATIONS")


class Typer:
    """
    Smooth out the lumps of Python type manipulation
    """

    def __init__(self, *, example=None, type_=None, function=None):
        if function:
            # find function return type
            inspect.signature(function)
        elif example:
            self.type_ = type(example)
        else:
            self.type_ = type_

    def __getattr__(self, item):
        try:
            attribute_type = self.type_.__annotations__[item]
            return Typer(type_=attribute_type)
        except:
            pass

        desc = arg_spec(self.type_, item)
        if desc:
            return_type = desc.annotations.get("return")
            if return_type:
                return parse(return_type)
            return_type = ANNOTATIONS.get((self.type_, item))
            if return_type:
                return return_type

            logger.error(
                "expecting {{type}}.{{item}} to have annotated return type",
                type=self.type_.__name__,
                item=item,
            )
        logger.error(
            """expecting {{type}} to have attribute {{item|quote}} declared with a type annotation""",
            type=self.type_.__name__,
            item=item,
        )

    def __add__(self, other):
        if self.type_ is str or other.type_ is str:
            return Typer(type_=str)
        logger.error("not handled")

    def __call__(self, *args, **kwargs):
        spec = inspect.getfullargspec(self.type_)

    def __str__(self):
        return f"Typer(class={self.type_.__name__})"


class CallableTyper(Typer):
    """
    ASSUME THIS WILL BE CALLED, AND THIS IS THE TYPE RETURNED
    """

    def __init__(self, *, type_):
        self.type_ = type_

    def __call__(self, *args, **kwargs):
        return Typer(type_=self.type_)

    def __getattr__(self, item):
        spec = inspect.getmembers(self.type_)
        for k, m in spec:
            if k == item:
                inspect.ismethod(m)

    def __str__(self):
        return f"CallableTyper(return_type={self.type_.__name__})"


class LazyTyper(Typer):
    """
    PLACEHOLDER FOR STREAM ELEMENT TYPE, UNKNOWN DURING LAMBDA DEFINITION
    """

    def __init__(self, resolver=None):
        Typer.__init__(self)
        self._resolver: Typer = resolver or (lambda t: t)

    def __getattr__(self, item):
        def build(type_):
            return getattr(type_, item)

        return LazyTyper(build)

    def __call__(self, *args, **kwargs):
        def build(type_):
            return type_

        return LazyTyper(build)

    def __str__(self):
        return "LazyTyper()"


export("mo_streams.type_parser", Typer)
export("mo_streams.byte_stream", Typer)



# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from mo_files import File
from mo_future import get_function_name
from mo_imports import export

from mo_streams._utils import ByteStream

DECODERS = {
    "zst": ByteStream.from_zst,
    "tar": ByteStream.from_tar,
    "zip": ByteStream.from_zip,
    # "gzip": gzip_stream
}


def _get_file_stream(file, stream):
    """
    RETURN A STREAM FROM THE GIVEN FILE
    :param file:
    :return:
    """
    name, extension = _get_extension(file)
    decoder = DECODERS.get(extension, None)
    if not decoder:
        return stream
    return _get_file_stream(name, decoder(stream))


def _extend(cls):
    """
    DECORATOR TO ADD METHODS TO CLASSES
    :param cls: THE CLASS TO ADD THE METHOD TO
    :return:
    """

    def extender(func):
        setattr(cls, get_function_name(func), func)
        return func

    return extender


@_extend(File)
def content(self):
    return _get_file_stream(self.os_path, ByteStream(open(self.os_path, "rb")))


@_extend(File)
def bytes(self):
    return ByteStream(open(self.os_path, "rb"))


def _get_extension(file_name):
    parts = file_name.split(".")
    if len(parts) > 1:
        name = ".".join(parts[:-1])
        extension = parts[-1]
        return name, extension
    return file_name, ""


class File_usingStream:
    """
    A File USING A BORROW STREAM.  FOR USE IN TAR AND ZIP FILES
    """

    rel_path: str

    def __init__(self, rel_path, content):
        self.rel_path = rel_path
        self._content = content

    def content(self):
        return self._content()

    def bytes(self):
        return self._content()


export("mo_streams.byte_stream", File_usingStream)


# encoding: utf-8
# THIS FILE IS AUTOGENERATED!
from __future__ import unicode_literals
from setuptools import setup
setup(
    author='Kyle Lahnakoski',
    author_email='kyle@lahnakoski.com',
    classifiers=["Development Status :: 5 - Production/Stable","Topic :: Software Development :: Libraries","Topic :: Software Development :: Libraries :: Python Modules","License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)","Programming Language :: Python :: 3.7","Programming Language :: Python :: 3.8","Programming Language :: Python :: 3.9"],
    description='More Streams! Chained function calls',
    extras_require={"tests":["mo-files","mo-logs","zstandard","boto3","moto","pandas"]},
    include_package_data=True,
    install_requires=["mo-dots==9.278.22339","mo-files==6.278.22339","mo-future==6.265.22338","mo-json==6.278.22339"],
    license='MPL 2.0',
    long_description='# More Streams!!\n\nPython code is more elegant with method chaining!\n\n\n## Overview\n\nThere are two families of "streams" in this library, both are lazy:\n\n1. `ByteStream` - a traditional stream of bytes intended to pipe bytes through various byte transformers, like compression, encoding and encyrption.  \n2. `ObjectStream`: An iterator/generator with a number of useful methods.\n\n### Example\n\nIn this case I am iterating through all files in a tar and parsing them:\n\n    results = (\n        File("tests/so_queries/so_queries.tar.zst")\n        .content()\n        .content()\n        .exists()\n        .utf8()\n        .to_str()\n        .map(parse)\n        .to_list()\n    )\n    \n Each of the steps constructs a generator, and no work is done until the last step\n \n \n * `File().content()` - will unzst and untar the file content to an `ObjectStream` of file-like objects.  It is short form for `stream(File().read_bytes()).from_zst().from_tar()`\n * The second `.content()` is applied to each of the file-like objects, returning `ByteStream` of the content for each\n * `.exists()` - some of the files (aka directories) in the tar file do not have content, we only include content that exists.\n * `.utf8` - convert to a `StringStream`\n * `.to_str` - convert to a Python `str`, we trust the content is not too large\n * `.map(parse)` - run the parser on each string\n * `.to_list()` - a "terminator", which executes the chain and returns a Python `list` with the results\n \n## Project Status\n\nAlive and in use, but \n\n* basic functions missing\n* inefficient - written using generators\n* generators not properly closed\n\n\n## Optional Reading\n\nThe method chaining style has two distinct benefits\n\n* functions are in the order they are applied \n* intermediate values need no temporary variables\n\nThe detriments are the same that we find in any declarative language: Incorrect code can be difficult to debug because you can not step through it to isolate the problem.  For this reason, the majority of the code in this library is dedicated to validating the links in the function chain before they are run.\n\n### Lessons\n\nThe function chaining style, called "streams" in Java or "linq" in C#, leans heavly on the strict typed nature of those langauges.  This is missing in Python, but type annotations help support this style of programming.\n\n',
    long_description_content_type='text/markdown',
    name='mo-streams',
    packages=["mo_streams"],
    url='https://github.com/klahnakoski/mo-streams',
    version='1.278.22339',
    zip_safe=False
)


# encoding: utf-8
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Contact: Kyle Lahnakoski (kyle@lahnakoski.com)
#

from mo_imports import export

from mo_json import JxType
from mo_streams._utils import Reader, Stream
from mo_streams.byte_stream import ByteStream
from mo_streams.object_stream import ObjectStream
from mo_streams.type_utils import Typer


class StringStream(Stream):
    def __init__(self, chunks):
        self._chunks = chunks

    def __getattr__(self, item):
        def read():
            for v in self._chunks:
                yield getattr(v, item), {}

        return ObjectStream(read(), getattr(Typer(type_=str), item), JxType())

    def utf8(self) -> ByteStream:
        return ByteStream(Reader((c.encode("utf8") for c in self._chunks)))

    def lines(self):
        def read():
            line = ""
            end = -1
            try:
                while True:
                    while end == -1:
                        content = next(self._chunks)
                        line += content
                        end = line.find("\n")

                    while end != -1:
                        yield line[:end], {}
                        line = line[end:]
                        end = line.find("\n")
            except StopIteration:
                if line:
                    yield line, {}

        return ObjectStream(read, Typer(type_=str), JxType())

    def to_str(self) -> str:
        return "".join(self._chunks)


export("mo_streams.byte_stream", StringStream)
